/* --------------------------------------------------------------------
 * 
 * Clac - calculator (C) 1994-2006 by Eero Tamminen.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms specified in the COPYRIGHTS file.
 *
 * Edit Makefile and type 'make readline' (or plain 'make') to compile.
 * You should have GNU Make.
 *
 * Clac: Calculates the expression and returns the result
 *       or error code in the packet structure.
 *
 * --------------------------------------------------------------------
 */

#include <ctype.h>         /* toupper()          */
#include <stdlib.h>        /* atof()             */
#include <string.h>        /* strcpy(), strcmp() */
#include <limits.h>        /* max values         */

#define CLAC_C
#include "clac.h"          /* constants & communication struct */
#include "function.h"      /* how to calculate functions       */
#undef  CLAC_C

extern int cplx_error;

/* --------------------------------------------------------------------
 * define some preprosessor symbols
 */

/* define internal allocation sizes (should be enough)   */
#define  DEPTH  64        /* max. parenth. nesting depth */
#define STACK  128        /* size of the operator stack  */
#define V_STACK  256      /* size of the value stack     */

/* --------------------------------------------------------------------
 * all structures are global
 */

/* globals + function identifier stack(s) */
static struct
{
  int error;         /* global error code */
  int v_valid;       /* value validation  */
  char *fname;       /* function names    */
} id = {0, 0, FUNCTIONS};

/* parenthesis and function stacks */
static struct
{
  int idx;               /* parenthesis level         */
  int max;               /* maximum idx               */
  char func;             /* current function opcode   */
  int opx[DEPTH + 1];    /* current op index for par  */
  int vax[DEPTH + 1];    /* current val index for par */
  char buf[DEPTH + 1];   /* function opcode buffer    */
} par = {0, DEPTH, '\0', {0}, {0}, ""};

/* operator stack  */
static struct
{
  int idx;
  int max;
  char buf[STACK + 1];
} op = {0, STACK, ""};

/* value stack  */
static struct value_stk {
  int idx;
  int max;
  cplx buf[V_STACK + 1];
} val = {0, V_STACK, {{0.0, 0.0}}};

/* --------------------------------------------------------------------
 * macros
 */

/* increment stack index and put value on stack (ie. PUSH) */
#define PUSH(stk,val)  if((stk).idx < (stk).max) {     \
                         (stk).idx += 1;               \
                         (stk).buf[(stk).idx] = (val); \
                       } else                          \
                         id.error = STK_ERR;

/* --------------------------------------------------------------------
 * declare subfunctions
 */

/* parser variable / function */
static cplx  get_predef(char *exp, int *offset, struct packet *p);
/* get variable's value */
static cplx  get_var(char *var, struct packet *p);
/* identify a function */
static void  get_func(char *name);
/* parse ascii & convert it to a number */
static double  asc2num(char *exp, int *offset);
/* parse a decimal from an exp. */
static double  get_dec(char *exp, int *offset);
/* parse value */
static double  get_value(char *exp, int *offset, int bits);
/* 'value' of a char */
static int  chr_pos(char c, char *base, int base_len);

/* parse prefix and in-between operations */
static void  operation(cplx value, char op);
/* parse unary operators */
static void  unary(char op);
/* apply a prefix to a value */
static void apply_prefix(void);
/* juggle stacks, if possible */
static void  eval_stack(void);
/* operator -> operator level */
static int  get_level(int stk_offset);
/* evaluate operation */
static cplx  apply(char op, cplx x, cplx y);
/* &, | operations */
static double  binops(int op, double x, double y);
/* >>, << operations */
static double  shiftops(int op, double x, double y);

/* increase parenthesis level */
static void  open_bracket(void);
/* decrease parenthesis level */
static cplx  close_bracket(cplx x, int flag, int mode );


void calc(struct packet *pkt)  /* evaluate one whole expression  */
{
  /* exp   : expression to evaluate          */
  /* value : current parsed value            */
  /* mark  : current character in expression */
  /* valid : expression validation flag, set when number parsed  */
  /* exit  : 'expression end' flag           */
  /* offset: character offset in expression  */

  int exit = FALSE, offset = 0;
  cplx value;
  char mark;

  /* Uses global variables:  */

  par.idx = 0;                       /* parenthesis stack pointer */
  par.opx[0] = par.vax[0] = 0;       /* additional stack pointers */
  par.func = op.idx = val.idx = -1;  /* no function, empty stack  */

  id.error = FALSE;
  cplx_error = FALSE;
  id.v_valid = FALSE;    /* value validation    */
  value.real = 0.0;
  value.imag = 0.0;

  /* parsering loop, repeated until expression ends */
  do
  {
    mark = pkt->exp[offset];
    switch(mark)
    {
      case ' ':
      case '\t':
        offset ++;    /* jump over spaces  */
        break;
      case '~':	      /* prefixes */
        unary(mark);
        offset++;
        break;
      case '>':        /* operators  */
      case '<':
      case '|':
      case '&':
      case '+':
      case '-':
      case '%':
      case '*':
      case '/':
      case '^':
        operation(value, mark);
        offset ++;
        break;
      case '(':
        open_bracket();
        offset ++;
        break;
      case ')':
        value = close_bracket(value, TRUE, pkt->mode_trig);
        offset ++;
        break;
      case ',':      /* more parameters  */
        if(par.idx > 0 && par.buf[par.idx - 1] >= MFUNC)
        {
          /* evaluate previous parameter    */
          value = close_bracket(value, FALSE, pkt->mode_trig);
          id.v_valid = FALSE;
          PUSH(val, value);
          open_bracket();  /* next param.  */
        }
        else
          pkt->error = FNC_ERR;
        offset ++;
        break;
      case '0':
        /* C notation for hex, or normal decimals? */
        if(pkt->exp[offset + 1] == 'x')
        {
          offset += 2;
          value.real = get_value(pkt->exp, &offset, 4);
          value.imag = 0.0;
          break;
        }
      case '1':        /* decimal  */
      case '2':
      case '3':
      case '4':
      case '5':
      case '6':
      case '7':
      case '8':
      case '9':
      case '.':
        value.real = get_dec(pkt->exp, &offset);
        value.imag = 0.0;
        break;
      case BIN_SYM:      /* binary decimal  */
        offset ++;
        value.real = get_value(pkt->exp, &offset, 1);
        value.imag = 0.0;
        break;
      case OCT_SYM:      /* octal decimal  */
        offset ++;
        value.real = get_value(pkt->exp, &offset, 3);
        value.imag = 0.0;
        break;
      case HEX_SYM:      /* hexadecimal    */
        offset ++;
        value.real = get_value(pkt->exp, &offset, 4);
        value.imag = 0.0;
        break;
      case '\"':
        offset++;
        value.real = asc2num(pkt->exp, &offset);
        value.imag = 0.0;
        break;

      default:
        /* a comment or an end of expression */
        if(mark == ';' || mark < ' ')
        {
          exit = TRUE;
          break;
        }
        /* a postfix  */
        if(id.v_valid)
        {
          switch(mark)
          {
            case '!':    /* factor  */
              value.real = factor(value.real, &id.error);
              value.imag = 0.0;
              break;
            case 'i':    /* imaginary  */
              value.imag = value.real;
              value.real = 0.0;
              break;
            default:
              id.error = GEN_ERR;
          }
          offset ++;
        }
        /* function or variable?  */
        else
          value = get_predef(pkt->exp, &offset, pkt);
      }
  /* until exit or error message */
  } while((exit == FALSE) && (id.error == FALSE));

  /* result of evaluation */
  if(val.idx >= 0)
    pkt->result = val.buf[val.idx];

  /* something to return? */
  if(!id.error)
  {
    if(id.v_valid == TRUE)
    {

      /* evaluate rest of the pkt->expression */
      operation( value, '|');
      /* mismatch error or result */
      if(par.idx)
        id.error = PAR_ERR;
      else
        pkt->result = val.buf[val.idx];

    }
    else
    {
      if((val.idx < 0) && (op.idx < 0))
      {
        /* no expr. */
        id.error = NO_EXP;
        pkt->result.real = 0.0;
        pkt->result.imag = 0.0;
      }
      /* trailing operators  */
      else
        id.error = GEN_ERR;
    }
  }

  if(id.error)
  {
    pkt->error = id.error;
    pkt->err_pos = offset;
  }
  else
    pkt->error = pkt->err_pos = 0;

  if(cplx_error)
    pkt->error = cplx_error;
}


/* ====================================================================
 *    parser variables & functions
 * ====================================================================
 */
static cplx get_predef(char *exp, int *offset, struct packet *pkt)
{
  /* return value of the parsed variable / make new variable
   * / assign new value to a variable / parser a function
   */

  int more = 0;
  char mark;
  cplx value;
  char str[MAX_VAR + 1];      /* variable name  */

  if(id.v_valid == FALSE)
  { /* preceded by operator?
     * parse one identification from expression
     */
    mark = toupper(exp[*offset]);
    while((((mark >= '0') && (mark <= '9'))
       || ((mark >= 'A') && (mark <= 'Z'))) && (more <= MAX_VAR))
    {
      str[more ++] = mark;
      mark = toupper(exp[*offset + more]);
    }
    /* not valid  */
    if(more == FALSE)
      id.error = GEN_ERR;
    else
    {
      /* too long */
      if(more > MAX_VAR)
        id.error = VAR_ERR;
      else
      {
        /* end name  */
        str[more] = '\0';

        /* check if parenthesis is separated  */
        while(exp[*offset + more] == ' ') more ++;
        mark = exp[*offset + more];

        /* a function */
        if(mark == '(')
          get_func(str);
        else
        { /* a variable  */
          value = get_var(str, pkt);
          id.v_valid = TRUE;
        }
      }
    }
  }
  else
    id.error = GEN_ERR;

  /* overtake identifier */
  *offset += more;
  return(value);
}

/* --------------------------------------------------------------------
 * find out value of a variable
 */
static cplx get_var(char *name, struct packet *pkt)
{
  int lenny, idx = 0;
  char *var;
  cplx value;

  /* find the index for the given variable */
  var = pkt->vname;
  lenny = strlen(var);
  while(lenny && (strcmp(var, name)))
  {
    var += ++lenny;
    lenny = strlen(var);
    idx ++;
  }

/* variable found */
  if(lenny > 0)
  {
    value.real = pkt->value[idx].real;
    value.imag = pkt->value[idx].imag;
  }
  else
    id.error = VAR_ERR;

  return(value);
}

/* ====================================================================
 *      parser a function
 * ====================================================================
 */
/* identify a function -> id to stack */
static void get_func(char *name)
{
  /* returns the character identifier of the parsed function */
  int idx = 0;
  char *fun;

  /* find the index for the given function */
  fun = id.fname;
  while(fun[0] && strcmp(fun, name))
  {
    fun += strlen(fun) + 1;
    idx ++;
  }

  /* function found */
  if(fun[0] > 0)
    par.func = idx;      /* function opcode  */
  else
    id.error = FNC_ERR;
}

/* ====================================================================
 *      parse a value
 * ====================================================================
 */
/* parse ascii  */
static double asc2num(char *exp, int *offset)
{
  double value = 0.0;

  if(id.v_valid == FALSE)
  {
    id.v_valid = TRUE;
    while(exp[(*offset)] > ' ')
    {
      value = value * 256.0 + (double)exp[*offset];
      (*offset)++;
    }
  }
  else
    id.error = GEN_ERR;

  return(value);
}

/* parse a decimal number */
static double get_dec(char *exp, int *offset)
{
  register char mark;
  int mark_set = FALSE, exp_set = FALSE;
  double value = 0.0;

  if(id.v_valid == FALSE)
  {
    id.v_valid = TRUE;
    value = atof(&exp[*offset]);
    /* jump over number */
    do
    {
      mark = exp[++(*offset)];
      /* check for multiple decimal points */
      if(mark == '.')
      {
        if(mark_set)
          id.error = GEN_ERR;
        else
          mark_set = TRUE;
      }
      /* check for multiple exponents */
      if(mark == 'e' || mark == 'E')
      {
        if(exp_set)
	{
          id.error = GEN_ERR;
	}
        else
        {
          /* check for exponent validity */
          mark = exp[++(*offset)];
          if(mark == '+' || mark == '-' ||
            (mark >= '0' && mark <= '9'))
          {
            mark_set = TRUE;
            exp_set = TRUE;
            mark = '.';
          }
          else
            id.error = GEN_ERR;
        }
      }
    } while(!id.error && ((mark >= '0' && mark <= '9') || mark == '.'));
  }
  else
    id.error = GEN_ERR;

  return(value);
}

/* --------------------------------------------------------------------
 * parsering for 2^bits number base(up to hex, at the moment)
 */
static double get_value(char *exp, int *offset, int bits)
{
  /* returns parsed value, changes expression offset    */

  double value = 0.0;
  int i, end, lenny, pos, index, len_long = sizeof(long) * 8;
  unsigned long num1 = 0, num2 = 0; /* int/decimal  parts  */
  char digit;
  char *base = "0123456789ABCDEF";  /* number base(s)      */
  end = 1 << bits;                  /* end of current base */
  if(bits == 1) len_long --;        /* eliminate negate    */
  lenny = len_long / bits;          /* max. number lenght  */

  /* if start of expression or preceded by an operator */
  if(id.v_valid == FALSE)
  {
    id.v_valid = TRUE;
    i = 0;
    digit = exp[*offset];
    index = chr_pos(digit, base, end);  /* digit value  */

    /* increment i until the integer part of value ends */
    while((i < lenny) && (index >= 0))
    {
      num1 = (num1 << bits) | index;
      digit = exp[++(*offset)];
      index = chr_pos(digit, base, end);
      i ++;
    }
    /* too long number or expands into the sign bit?  */
    if(((i == lenny) && (index >= 0)) || (num1 & (1L << (len_long - 1))))
      id.error = OVR_ERR;
    else
    {
      /* decimal part? */
      if(digit == '.')
      {
        pos = len_long - bits;
        digit = exp[++(*offset)];
        index = chr_pos(digit, base, end);

        /* calculate x / 0xFFFFFFFF */
        while((pos >= bits) && (index >= 0))
        {
          pos -= bits;
          num2 |= (long)index << pos;
          digit = exp[++(*offset)];
          index = chr_pos(digit, base, end);
        }
        /* jump over any remaining decimals  */
        if((pos < bits) && (index >= 0))
          while(chr_pos(exp[*offset], base, end) >= 0)
            (*offset) ++;
      }
      /* compose value of integral and decimal parts  */
      value = num2 / (double) (1L << (len_long - bits)) + num1;
    }
  }
  else
    id.error = GEN_ERR;

  return(value);
}

/* --------------------------------------------------------------------
 * return the value of a given character in the current number base
 */

static int chr_pos(char chr, char *string, int len)
{
  /* returns a character position (0 - ) in a string or -1  */
  int pos = 0;        /* character position */
  chr = toupper(chr); /* uppercase a char   */

  /* till strings end or character found */
  while(pos < len && *(string ++) != chr)
    pos ++;

  /* if string end -> not found */
  if(pos == len)
    return(-1);
  else
    return(pos);
}

/* ====================================================================
 *      expression evaluation
 * ====================================================================
 */
static void operation(cplx value, char oper)
{
  /* uses globals par[], id.error[], op[], val[]
   * operation executed if the next one is on same or lower level
   */
  /* something to calc? */
  if(id.v_valid == TRUE)
  {
    /* next number */
    id.v_valid = FALSE;

    /* add new items to stack */
    PUSH(op, oper);
    PUSH(val, value);

    /* more than 1 operator  */
    if(op.idx > par.opx[par.idx])
    {
      /* but only one value */
      if(val.idx == par.vax[par.idx])
        apply_prefix();
      else
        /* evaluate all possible operations */
        eval_stack();
    }
  }
  else
    unary(oper);
}

/* unary operators */
static void unary(char oper)
{
  /* check pre-value operators
   * have to be parenthesised
   */
  if(id.v_valid == FALSE && op.idx < par.opx[par.idx])
  {
    switch(oper)
    {
      case '+':		/* not needed */
        break;
      case '-':
      case '~':
        PUSH(op, oper);
        break;
      default:
        id.error = PRG_ERR;
    }
  }
  else
    id.error = GEN_ERR;
}

/* apply a prefix to the current value */
static void apply_prefix(void)
{
  cplx *value;

  value = &val.buf[val.idx];
  op.idx--;

  switch(op.buf[op.idx])
  {
    case '-':
      value->real = -value->real;
      value->imag = -value->imag;
      break;
    case '~':
      value->real = -value->real - 1;	/* bitwise not */
      value->imag = 0.0;
      break;
    default:
      id.error = PRG_ERR;
  }
  op.buf[op.idx] = op.buf[op.idx + 1];
}

/* --------------------------------------------------------------------
 * evaluate operators which precedence allows
 */
static void eval_stack()
{
  /* uses globals par[], op[], val[] */

  /* # of operators >= 2 and prev. op-level >= current op-level? */
  while((op.idx > par.opx[par.idx]) && get_level(-1) >= get_level(0))
  {
    /* shorten value stacks by one
     * + calculate resulting value
     */
    op.idx--;
    val.idx--;
    val.buf[val.idx] = apply(op.buf[op.idx],
      val.buf[val.idx], val.buf[val.idx + 1]);

    /* pull the just used operator out of the stack */
    op.buf[op.idx] = op.buf[op.idx + 1];
  }
}

/* --------------------------------------------------------------------
 * return the precedence level of a given operator
 */
static int get_level(int offset)
{
  /* used globals par[], op[]
   * returns operator level of: operator[stack index + offset]
   */
  switch(op.buf[op.idx + offset])
  {
    case '|':      /* binary operations  */
    case '&':
      return(0);

    case '>':      /* bit shifting    */
    case '<':
      return(1);

    case '+':
    case '-':
      return(2);

    case '%':      /* modulo    */
    case '*':
    case '/':
      return(3);

    case '^':      /* power */
      return(4);

    default:
      id.error = PRG_ERR;
  }
  return(6);
}

/* --------------------------------------------------------------------
 * apply operator to given values, return the result
 */
static cplx apply(char opcode, cplx value1, cplx value2)
{
  /* uses global id.error[]
   * returns the result of operation
   */
  /* more accurate (+, -, *, /) with less calculation */
  if(value1.imag == 0.0 && value2.imag == 0.0)
    {
    switch(opcode)
    {
        case '|':
        case '&':
          value1.real = binops(opcode, value1.real, value2.real);
          break;
        case '>':
        case '<':
          value1.real = shiftops(opcode, value1.real, value2.real);
          break;
        case '+':
          value1.real += value2.real;
          break;
        case '-':
          value1.real -= value2.real;
          break;
        case '%':
          if(value2.real != 0.0) 
	  {
            if(value1.real < 0.0)
              value1.real -= value2.real * ceil(value1.real / value2.real);
            else
              value1.real -= value2.real * floor(value1.real / value2.real);
	  }
          break;
        case '*':
          value1.real *= value2.real;
          break;
        case '/':
          /* not 'divide by zero'  */
          if(value2.real != 0.0)
            value1.real /= value2.real;
          else
            id.error = DEF_ERR;
          break;
        case '^':
          value1.real = pow(value1.real, value2.real);
          break;
        default:
          id.error = PRG_ERR;
      }
  }
  else
  {
    switch(opcode)
    {
      case '|':
      case '&':
        value1.real = binops(opcode, value1.real, value2.real);
        value1.imag = 0.0;
        break;
      case '>':
      case '<':
        value1.real = shiftops(opcode, value1.real, value2.real);
        value1.imag = 0.0;
        break;
      case '+':
        value1 = cplx_add(value1, value2);
        break;
      case '-':
        value1 = cplx_sub(value1, value2);
        break;
      case '%':
        /* real values only... */
        if(value2.real != 0.0)
	{
          if(value1.real < 0.0)
            value1.real -= value2.real * ceil(value1.real / value2.real);
          else
            value1.real -= value2.real * floor(value1.real / value2.real);
	}
        value1.imag = 0.0;
        break;
      case '*':
        value1 = cplx_mul(value1, value2);
        break;
      case '/':
        /* not 'divide by zero'  */
        if(value2.real == 0.0 && value2.imag == 0.0)
          id.error = DEF_ERR;
        else
          value1 = cplx_div(value1, value2);
        break;
      case '^':
        /* complex to real power only, sorry ;-)  */
        value1 = cplx_pow(value1, value2.real);
        break;
      default:
        id.error = PRG_ERR;
    }
  }
  /* Out of range... */
  if(value1.real >= HUGE_VAL || value1.imag >= HUGE_VAL)
    id.error = OVF_ERR;

  return(value1);        /* return result  */
}

static double binops(int op, double x, double y)
{
  double z;
  long xx, yy;

  xx = (long) x;
  yy = (long) y;

  /* in limits */
  if(xx > LONG_MAX || yy > LONG_MAX)
  {
    id.error = OVF_ERR;
    return(y);
  }
  else
  {
    z = (double) (op == '&' ? xx & yy : xx | yy);

    /* operate on 16 bits after the decimal point too */
    xx = (long) ((x - (double) xx) * 65536.0);
    yy = (long) ((y - (double) yy) * 65536.0);

    z += (double) (op == '&' ? xx & yy : xx | yy) / 65536.0;
    return(z);
  }
}

static double shiftops(int op, double x, double y)
{
  long multiple = 1;

  if(sizeof(long) * 8 >= (int) y)
  {
    multiple <<= (int) y;
    if(op == '<')
      return(x * (double)multiple);
    else
      return(x / (double)multiple);
  }
  else
  {
    id.error = OVF_ERR;
    return(y);
  }
}

/* ====================================================================
 *      parenthesis and help
 * ====================================================================
 */
static void open_bracket()
{
  /* preceded by operator */
  if(id.v_valid == FALSE)
  {
    /* not nested too deep  */
    if(par.idx < DEPTH)
    {
      /* inc. level */
      PUSH(par, (char)par.func);
      par.opx[par.idx] = op.idx + 1;
      par.vax[par.idx] = val.idx + 1;

      if(par.func >= MFUNC)
      {
        par.func = -1;   /* no function  */
        open_bracket(); /* like avg((1),(1+1))  */
      }
      else
        par.func = -1;
    }
    else
      id.error = STK_ERR;
  }
  else
    id.error = GEN_ERR;
}

/* --------------------------------------------------------------------
 * close prenthesis, and evaluate if it was a function
 */
/* last parsed value, last param. flag, trigonometric mode */
static cplx close_bracket(cplx value, int straight, int tmode)
{
  /* returns the value of the parenthesised expression */
  int func;

  /* preceded by an value */
  if(id.v_valid == TRUE)
  {
    /* prenthesis has a pair */
    if(par.idx > 0)
    {
      /* calculate the value of parenthesised exp. */
      operation(value, '|');
      value = val.buf[val.idx];

      /* if function, evaluate it now
       * and change value accordingly
       */
      func = par.buf[par.idx];
      /* one parameter */
      if(func >= 0)
        value = eval_func(func, value, tmode, &id.error);
      else
      {
        /* multiple parameters? */
        func = par.buf[par.idx - 1];
        if(straight && (func >= MFUNC))
        {
          par.idx --;
          value = eval_mfunc(func, par.vax[par.idx], val.idx, &id.error, val.buf);
        }
      }
      op.idx = par.opx[par.idx] - 1;  /* restore prev  */
      val.idx = par.vax[par.idx] - 1;
      par.idx --;
      /* next operator */
      id.v_valid = TRUE;
    }
    else
      id.error = PAR_ERR;
  }
  else
    id.error = GEN_ERR;

  return(value);
}
