/* --------------------------------------------------------------------
 *
 * Clac - calculator (C) 1994-2006 by Eero Tamminen.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms specified in the COPYRIGHTS file.
 *
 * Should work with both 16 & 32 bit ints. Use tab size 8.
 *
 * Edit Makefile and type 'make readline' (or plain 'make') to compile.
 * You should have GNU Make.
 *
 * Front-end:  Takes care of the acquiring of the expressions,
 *     outputting results and managing variables.
 *
 * --------------------------------------------------------------------
 */

#include <stdio.h>
#include <ctype.h>   /* toupper()        */
#include <string.h>  /* strcpy/cmp/len() */

#define CLI_C
#include "clac.h"
#include "common.h"
#include "version.h"
#undef  CLI_C

/* --------------------------------------------------------------------
 * Here are some of the cli's commandline option flags.
 */
#define CL_EXP  0x01  /* there *is* an expression  */
#define USE_FL  0x02  /* input from a file         */
#define CL_ERR  0x03  /* commanline error          */
#define MAX_FL  256   /* max length for a filename */

/* --------------------------------------------------------------------
 * communication structure
 * error, err_pos, mode_trig, result, expression, var. names & values
 */
static struct packet comm = {FALSE, 0, RAD, {0.0, 0.0}, "", "PI\0E\0", {{M_PI, 0.0}, {M_E, 0.0}}};

/* global variables: ouput & complex modes */
static int mode_out = DEC;
static int mode_cplx = CPX_CMPNT;
static int verbose = TRUE;

/* --------------------------------------------------------------------
 * declare cli subfunctions
 */
static char expression[MAX_LN];

/* Read expressions using readline? */
#ifdef READLINE
#define HISTORY 32                          /* history buffer size  */
extern void add_history(const char *read);  /*** doesn't free()!!!  */
extern char *readline(const char *prompt);
static int history;
#else
char *readline(const char *prompt)
{
  printf(prompt);
  return(fgets(expression, MAX_LN, stdin));
}
#endif

/* parser commandline: get modes and filename / expression */
static int  parse(int argc, char *argv[], char **file, char **exp);
static void file_exps(FILE *fp);  /* get exp's from file     */
static void evaluate(void);       /* execute evaluation      */
static int  mode_change(char c);  /* change mode             */
static void help(char buf[]);     /* print a brief help text */

/* ====================================================================
 *        main
 * ====================================================================
 */
int main(int argc, char *argv[])
{
  /* cmd_line opts. & output mode */
  int options;
  char *file_in, *prompt = "";
  FILE *fp;

  /* execute ini-file if it exists */
  if((fp = fopen("cli.ini", "r")) != NULL)
    file_exps(fp);

  /* parser the commandline */
  options = parse(argc, argv, &file_in, &comm.exp);

  switch(options)
  {
    /* command line syntax error */
    case CL_ERR:
      break;
    /* evaluate a command line expression */
    case CL_EXP:
      evaluate();
      break;
    /* use input stream to get expressions */
    case USE_FL:
      if((fp = fopen(file_in, "r")) == NULL)
        puts("Input file not available.");
      else
        file_exps(fp);
      break;
    /* interactive mode */
    default:
      if(verbose)
      {
        puts(VERSION " '?' for help.");
        prompt = "Expression: ";
      }
      while((comm.exp = readline(prompt)))
      {
	if (comm.exp[0] < 32)
	  break;
#ifdef READLINE
        if(history < HISTORY)
        {
          add_history(comm.exp);
          history ++;
        }
#endif
        evaluate();    /* calculate expression  */
      }
  }
  return(0);
}

/* ====================================================================
 *      parse a commandline
 * ====================================================================
 */
static int parse(int argc, char *argv[], char **fname, char **exp)
{
  /* Check the validity of the command line and assign the
   * filename to fname_in,  expression to exp, output mode to
   * mode_out trigonometric mode to comm.mode_trig (global)
   * and complex ouptput mode to mode_cplx.
   */
  int i = 1;

  /* no options */
  if(argc < 2)
    return(FALSE);

  /* set modes */
  while(argc > i && (argv[i][0] == '-'))
  {
    if(argv[i][1] == 'q')
      verbose = FALSE;
    else
      if(mode_change(argv[i][1]))
        return(CL_ERR);
    i ++;
  }

  if(argc > i)
  {
    /* from a file */
    if(argv[i][0] == '@')
    {
      *fname = &argv[i][1];
      return(USE_FL);
    }
    *exp = expression;
    /* expression */
    strcpy(*exp, argv[i]);
    while(argc > ++ i)
      strcat(*exp, argv[i]);
    return(CL_EXP);
  }
  else
    return(FALSE);
}


/* ====================================================================
 *    expression parsering and evaluation
 * ====================================================================
 */
/* evaluate expressions from a file */
static void file_exps(FILE *fp)
{
  comm.exp = expression;
  while(fgets(comm.exp, MAX_LN, fp))
  {
    if(verbose)
      printf(comm.exp);
    /* something to calc? */
    if(comm.exp[0] >= ' ')
      evaluate();
  }
  fclose(fp);
}

/* -------------------------------------------------------------------- */
static void evaluate(void)
{
  int i = 0, loop = TRUE;
  char name[MAX_VAR + 1];                /* variable name   */
  char result[2 * sizeof(long) * 8 + 2]; /* longest is <8*binary>.<8*binary> */
  name[0] = '\0';
  comm.error = 0;

  /* check for non-expression things... */
  while(loop)
  {
    switch(comm.exp[i])
    {
      /* variable assignment? */
      case '=':
        comm.exp = var_get(i, name, comm.exp, &comm.error);
        break;
      /* help */
      case '?':
        comm.error = NO_EXP;
        help(result);
        break;
      /* mode change */
      case ':':
        mode_change(comm.exp[i + 1]);
        comm.error = NO_EXP;
        break;
      /* comment */
      case ';':
        loop = FALSE;
        break;
      default:
        /* end of expression */
        if(comm.exp[i] < ' ' && comm.exp[i] != '\t')
          loop = FALSE;
    }
    i ++;
  }

  /* CALCULATE */
  if(!comm.error)
    calc(&comm);

  if(!comm.error)
    /* result into a string according to output mode  */
    comm.error = conv(comm.result, result, mode_out, mode_cplx, comm.mode_trig);

  /* get error text */
  if(comm.error)
    errors(comm.error, result);
  else
  {
    /* variable assignment */
    if(name[0])
      var_put(name, &comm);
    if(comm.error)
      errors(comm.error, result);
  }

  /* show result or error */
  if(comm.error != NO_EXP)
    puts(result);
}

/* change mode */
static int mode_change(char mark)
{
  switch(toupper(mark))
  {
    case 'B':
      mode_out = BIN;
      break;
    case 'O':
      mode_out = OCT;
      break;
    case 'D':
      mode_out = DEC;
      break;
    case 'H':
      mode_out = HEX;
      break;
    case 'C':
      mode_cplx = CPX_CMPNT;
      break;
    case 'V':
      mode_cplx = VEC_CMPNT;
      break;
    case 'P':
      mode_cplx = VEC_POLAR;
      break;
    case 'R':
      comm.mode_trig = RAD;
      break;
    case 'G':
      comm.mode_trig = DEG;
      break;
    default:
      puts("Output base modes   : binary- (b), octal- (o), hexa- (h), decimal (d).");
      puts("Complex output modes: component (c), vector (v), polar (p).");
      puts("Trigonometric modes : radians (r), degrees (g).");
      return(CL_ERR);
  }
  return(0);
}

/* --------------------------------------------------------------------
 * show a short summary of the current features of clac + cli
 */
static void help(char buf[])
{
  int i, a = 0;

  puts(VERSION);
  puts("- Uses the order of precedence (can be changed with parenthesis).");
  puts("- Number bases: Decimal, $=Hexadecimal, #=Octal, '=Binary and \"=ascii.");
  puts("- Output modes: Decimal (default), binary, octal and hexadecimal.");
  puts("- Complex output modes: Component (default), vector and polar.");
  puts("- Trigonometric modes: Radians (default) and degrees.");
  puts("  Select modes with command line options (-) or mode operators (:)");
  puts("  d, b, o, h, c, v, p, r and g. Eg. in interactive mode following");
  puts("  mode changes would restore the default state: ':d', ':r', ':c'.");
  puts("- Imaginary numbers are postfixed with 'i' like in '1 + 2i'.");
  puts("- Variables: Variables are case-insensitive and assigned with '='.");
  puts("  There are two predefined variables: E and PI.");
  puts("- Operators in the decending order of precedence:");
  puts("             +, -, !, ^, *, /, %, +, -, >, <, &, |.");
  puts("- Functions: lg(), ln(), rad(), deg(), sin(), cos(), tan(),");
  puts("             asin(), acos(), atan(), sinh(), cosh(), tanh(),");
  puts("             sqrt(), root(), rnd(), abs(), int(), frac(),");
  puts("             real(), imag(), rtop(), ptor(), min(), max(),");
  puts("             avg(), std(), over(), fraction().");
  puts("- Examples: (('101010 & 0x88ff) + 1)^2");

  /* print variable names and values */
  for(i = 0; comm.vname[i] != '\0'; i += strlen(&comm.vname[i]) + 1)
  {
    conv(comm.value[a], buf, mode_out, mode_cplx, comm.mode_trig);
    printf("%s = %s\n", &comm.vname[i], buf);
    a ++;
  }
}
