/* ====================================================================	*/
/*			*** Clac engine ***				*/
/* ====================================================================	*/

#include <stdlib.h>				/* atof()		*/
#include <math.h>				/* pow()		*/
#include "bare.h"

/* all structures are global						*/

/* define internal allocation sizes (should be enough ;-)		*/
#define	DEPTH	64			/* max. parenth. nesting depth	*/
#define STACK	128			/* size of the operator stack	*/
#define V_STACK	128			/* size of the value stack	*/

/* globals + function identifier stack(s)				*/
struct {
	int error;			/* global error code		*/
	int valid;			/* value validation		*/
} id = {0, 0};

/* parenthesis and function stacks					*/
struct {
	int idx;			/* parenthesis level		*/
	int max;			/* maximum idx			*/
	int opx[DEPTH + 1];		/* current op index for par	*/
	int vax[DEPTH + 1];		/* current val index for par	*/
} par = {0, DEPTH, {0}, {0}};

struct {					/* operator stack	*/
	int idx;
	int max;
	char buf[STACK + 1];
} op = {0, STACK, ""};

struct value_stk {					/* value stack	*/
	int idx;
	int max;
	double buf[V_STACK + 1];
} val = {0, V_STACK, {0.0}};

/* -------------------------------------------------------------------- */
/* macros								*/

/* increment stack index and put value on stack (ie. PUSH)		*/
#define PUSH(stk,val)	if((stk).idx < (stk).max) {			\
				(stk).idx += 1;				\
				(stk).buf[(stk).idx] = (val);		\
			} else						\
				id.error = STK_ERR;

/* -------------------------------------------------------------------- */
/* declare subfunctions							*/

static double	get_dec (char exp[], int *offset);
/* parse in-between operations	*/
static void	operation(double value, char op);
/* parse unary operators	*/
static void	unary (char op);
/* juggle stacks, if possible	*/
static void	eval_stack(void);
/* operator -> operator level	*/
static int	get_level(int stk_offset);
/* evaluate operation		*/
static double	apply(char op, double x, double y);

/* increase parenthesis level	*/
static void	open_bracket(void);
/* decrease parenthesis level	*/
static double	close_bracket(double x);


int calc (char *in, struct clac_out *out)	/* evaluate expression	*/
{
	/* in	 : expression to evaluate				*/
	/* value : current parsed value					*/
	/* mark	 : current character in expression			*/
	/* valid : expression validation flag, set when number parsed	*/
	/* exit	 : 'expression end' flag				*/
	/* offset: character offset in expression			*/

	int exit = FALSE, offset = 0;
	double value;
	char mark;

	/* Uses global variables:	*/

	par.idx = 0;			/* parenthesis stack pointer	*/
	par.opx[0] = par.vax[0] = 0;	/* additional stack pointers	*/
	op.idx = val.idx = -1;

	id.error = FALSE;
	id.valid = FALSE;		/* value validation		*/
	value = 0.0;

	/* parsering loop, repeated until expression ends		*/
	do {
		mark = in[offset];
		switch(mark) {
		case ' ':
		case '\t':
			offset ++;		/* jump over spaces	*/
			break;
		case '+':
		case '-':
		case '*':
		case '/':
		case '^':
			operation (value, mark);
			id.valid = FALSE;
			offset ++;
			break;
		case '(':
			open_bracket ();
			offset ++;
			break;
		case ')':
			value = close_bracket (value);
			offset ++;
			break;
		case '.':				/* decimal	*/
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			value = get_dec (in, &offset);
			id.valid = TRUE;
			break;
		default:
			/* end of expression or error... */
			if(mark < ' ' || mark == ';')
				exit = TRUE;
			else
				id.error = GEN_ERR;
		}

	/* until exit or error message					*/
	} while((exit == FALSE) && (id.error == FALSE));

        /* result of evaluation 					*/
        if (val.idx >= 0)
        	out->result = val.buf[val.idx];

	/* something to return?						*/
	if (!id.error) {
		if (id.valid) {

			/* evaluate rest of the expression		*/
			operation ( value, '+');
			if (par.idx)			/* mismatched	*/
				id.error = PAR_ERR;
			else				/* result out	*/
				out->result = val.buf[val.idx];

		} else {
			if ((val.idx < 0) && (op.idx < 0)) {
				out->result = 0.0;
			} else			/* trailing operators	*/
				id.error = GEN_ERR;
		}
	}

	if (id.error) {
		out->error = id.error;
		return(id.error);
	} else {
		out->error = 0;
		return(0);
	}
}

/* ==================================================================== */
/*			parse a value					*/
/* ==================================================================== */

/* parse a decimal number	*/
static double get_dec (char exp[], int *offset)
{
	char mark;
	int off, mark_set = FALSE, exp_set = FALSE;
	double value = 0.0;

	if (id.valid == FALSE) {
		off = *offset;
		/* search number end and change ',' to '.'. */
		/* jump over number */
		do {
			if (exp[++(*offset)] == ',')
				exp[*offset] = '.';
			mark = exp[*offset];
			/* check for multiple decimal points */
			if (mark == '.')
				if (mark_set)
					id.error = GEN_ERR;
				else
					mark_set = TRUE;
			/* check for multiple exponents */
			if (mark == 'e' || mark == 'E')
				if (exp_set)
					id.error = GEN_ERR;
				else {
					/* check for exponent validity */
					mark = exp[++(*offset)];
					if (mark == '+' || mark == '-' ||
					   (mark >= '0' && mark <= '9')) {
						mark_set = TRUE;
						exp_set = TRUE;
						mark = '.';
					} else
						id.error = GEN_ERR;
				}
		} while (!id.error && ((mark >= '0' && mark <= '9') || mark == '.'));
		value = atof (&exp[off]);
	} else
		id.error = GEN_ERR;

	return (value);
}

/* ==================================================================== */
/*			expression evaluation				*/
/* ==================================================================== */

static void operation (double value, char oper)
{
	/* uses globals par[], id.error[], op[], val[]		*/
	/* operation executed if the next one is on same or lower level	*/

	if (id.valid) {				/* something to calc?	*/
		/* add new items to stack and increase pointer		*/
		PUSH (op, oper);
		PUSH (val, value);

		if (op.idx > par.opx[par.idx]) {	/* > 1 operator	*/

			if (val.idx == par.vax[par.idx]) {	/* sign	*/
				val.buf[val.idx] = -value;
				op.idx --;
				op.buf[op.idx] = op.buf[op.idx + 1];
			} else
				eval_stack ();
				/* evaluate all possible operations	*/
		}
	} else {
		/* pre- or post-operators instead of in-betweens	*/
		unary (oper);
	}
}

/* unary operators	*/
static void unary (char oper)
{
	/* check pre-value operators */
	switch (oper) {
	case '-':
	case '+':
		/* signs */
		if (op.idx < par.opx[par.idx]) {
			if (oper == '-')
				PUSH (op, oper);
		} else
			id.error = GEN_ERR;
		break;
	default:
		id.error = GEN_ERR;
	}
}

/* -------------------------------------------------------------------- */
/* evaluate operators if precedence allows it				*/

/* evaluate all possible (according to order of precedence) operators	*/
static void eval_stack ()
{
	/* uses globals par[], op[], val[]	*/

	/* # of operators >= 2 and prev. op-level >= current op-level ?	*/
	while ((op.idx > par.opx[par.idx]) && get_level (-1) >= get_level (0)) {

		/* shorten value stacks by one	*/
		/* + calculate resulting value	*/
		op.idx -= 1;
		val.idx -= 1;
		val.buf[val.idx] = apply(op.buf[op.idx],
			val.buf[val.idx], val.buf[val.idx + 1]);

		/* pull the just used operator out of the stack		*/
		op.buf[op.idx] = op.buf[op.idx + 1];
	}
}

/* -------------------------------------------------------------------- */
/* return the precedence level of a given operator			*/

static int get_level (int offset)
{
	/* used globals par[], op[]					*/
	/* returns operator level of: operator[stack index + offset]	*/

	switch (op.buf[op.idx + offset]) {
		case '+':
		case '-':
			return (0);
			break;
		case '*':
		case '/':
			return (1);
			break;
		case '^':
			return (2);
			break;
	}
	return (6);
}

/* -------------------------------------------------------------------- */
/* apply operator to given values, return the result			*/

static double apply (char opcode, double value1, double value2)
{
	/* uses global id.error[]			*/
	/* returns the result of operation	*/

	switch (opcode) {
		case '+':
			value1 += value2;
			break;
		case '-':
			value1 -= value2;
			break;
		case '*':
			value1 *= value2;
			break;
		case '/':
			/* not 'divide by zero'	*/
			if (value2 != 0.0)
				value1 /= value2;
			else
				id.error = DEF_ERR;
			break;
		case '^':
			value1 = pow(value1, value2);
			break;
	}
	return(value1);				/* return result	*/
}

/* ==================================================================== */
/*			parenthesis and help				*/
/* ==================================================================== */

static void open_bracket ()
{
	if (id.valid == FALSE) {	/* preceded by operator		*/
		if (par.idx < DEPTH) {	/* not nested too deep		*/
			par.idx ++;
			par.opx[par.idx] = op.idx + 1;
			par.vax[par.idx] = val.idx + 1;
		} else
			id.error = STK_ERR;
	} else
		id.error = GEN_ERR;
}

/* -------------------------------------------------------------------- */
/* close prenthesis, and evaluate					*/

/* last parsed value, last param. flag, trigonometric mode	*/
static double close_bracket (double value)
{
	/* returns the value of the parenthesised expression	*/

	if (id.valid) {			/* preceded by an operator	*/
		if (par.idx > 0) {	/* prenthesis has a pair	*/
			/* calculate the value of parenthesised exp.	*/
			operation (value, '+');
			value = val.buf[val.idx];
			op.idx = par.opx[par.idx] - 1;	/* restore prev	*/
			val.idx = par.vax[par.idx] - 1;
			par.idx --;
		} else
			id.error = PAR_ERR;
	} else
			id.error = GEN_ERR;

	return (value);
}
