/* ------------------------------------------------------------ 
 * Clac - calculator (C) 1994-2006 by Eero Tamminen.
 *
 * This is free software; you can redistribute it and/or modify
 * it under the terms specified in the COPYRIGHTS file.
 * 
 * calculate functions parsed by clac()
 */

#include <limits.h>
#define FUNCTION_C
#include "clac.h"
#include "function.h"
#undef  FUNCTION_C

/* maybe I'll imagine something fancier for all these functions someday
 * ---------------------------------------------------------------------
 */
static cplx tofraction(cplx stk[]);

/* return the factor of x */
double factor(double x, int *err)
{
  int fac;

  fac = (int) x;
  if(fac > 0 && x == (double) fac)
    for(fac --; (fac > 0 && x < HUGE_VAL); fac --)
      x *= (double) fac;
  else
    *err = DEF_ERR;

  if(x >= HUGE_VAL)
    *err = OVF_ERR;
  return x;
}

/* -----------------------------------------------------------
 * evaluate function on the given parameter
 */

/* function opcode, value for function, triggonometric mode */
cplx eval_func(char code, cplx value, int tmode, int *err)
{
  /* uses global id->error[]
   * returns the result of a function
   */

  /* check argument for things that can't be evaluated.
   * logarithms, square root, arcus...
   */
  if(((code == FN_LN || code == FN_LG) && value.real <= 0.0) ||
     (code == FN_SQRT && value.real < 0.0) ||
     ((code == FN_ASIN || code == FN_ACOS || code == FN_ATAN) &&
      (value.real < -1.0 || value.real > 1.0)))
  {
    *err = DEF_ERR;
    return(value);
  }

  /* In case function is trigonometric and angle mode is degrees.
   * trusts the functions to be in certain order...
   */
  if(code >= FN_SIN && code <= FN_ATAN && tmode == DEG)
      value.real = TO_RAD(value.real);

  /* Opcode refers to the fuction name position in id->fname in clac.c */
  switch(code)
  {
    case FN_LN:
      value.real = log(value.real);
      break;
    case FN_LG:
      value.real = log10(value.real);
      break;
    case FN_RAD:           /* degrees <-> radians  */
      value.real = TO_RAD(value.real);
      break;
    case FN_DEG:
      value.real = TO_DEG(value.real);
      break;
    case FN_SIN:           /* trigonometric functions  */
      value.real = sin(value.real);
      break;
    case FN_COS:
      value.real = cos(value.real);
      break;
    case FN_TAN:
      value.real = tan(value.real);
      break;
    case FN_SINH:          /* hyperbolic -  */
      value.real = sinh(value.real);
      break;
    case FN_COSH:
      value.real = cosh(value.real);
      break;
    case FN_TANH:
      value.real = tanh(value.real);
      break;
    case FN_ASIN:          /* arcus -  */
      value.real = asin(value.real);
      break;
    case FN_ACOS:
      value.real = acos(value.real);
      break;
    case FN_ATAN:
      value.real = atan(value.real);
      break;
    case FN_SQRT:          /* others  */
      value.real = sqrt(value.real);
      break;
    case FN_RND:           /* random (x)  */
      value.real = rand() * value.real / (double) RAND_MAX;
      break;
    case FN_ABS:           /* absolute (x)  */
      value.real = fabs(value.real);
      break;
    case FN_INT:           /* int part  */
      value.real = (value.real > 0.0 ? (floor(value.real)) : (ceil(value.real)));
      break;
    case FN_FRAC:          /* frac part  */
      value.real = fabs(value.real) - floor(value.real);
      break;
    case FN_REAL:          /* real part  */
      value.imag = 0.0;
      break;
    case FN_IMAG:          /* imag part  */
      value.real = 0.0;
      break;
    default:
      *err = PRG_ERR;
  }

  /* In case trigonometric function is arcus...+ mode is degrees.
   * trusts functions enums to be in certain order
   */
  if(code >= FN_ASIN && code < FN_ATAN && tmode == DEG)
      value.real = TO_DEG(value.real);

  if(value.real >= HUGE_VAL || value.imag >= HUGE_VAL)
    *err = OVF_ERR;

  return value;
}

/* --------------------------------------------------------------------
 * evaluate function on multiple parameters (no math Warp Mode, sniff)
 */

/* function opcode, start/end of params., value stack      */
cplx eval_mfunc(char code, int a, int b, int *err, cplx stack[])
{
  int i, params;
  cplx value;
  double tmp;
  value.real = 0.0;
  value.imag = 0.0;
  params = b - a + 1;

  switch(code)
  {
    case FN_RTOP:                         /* rectancular to polar  */
    case FN_PTOR:                         /* polar to rectangular  */
      switch(params)
      {
        case 1:                             /* Complex.  */
          value.real = stack[a].real;
          value.imag = stack[a].imag;
          value = change_type(value, (code == 1 ? POLAR : RECT));
          break;
        case 2:                             /* Two doubles.  */
          value.real = stack[a].real;
          value.imag = TO_RAD(stack[b].real);
          value = change_type(value, (code == 1 ? POLAR : RECT));
          break;
        default:
          *err = ARG_ERR;
      }
      break;

    case FN_MIN:                         /* smallest value  */
      value.real = stack[a].real;
      for(i = a; i < b;)
        if(stack[++ i].real < value.real)
          value.real = stack[i].real;
      break;
    case FN_MAX:                         /* largest value  */
      value.real = stack[a].real;
      for(i = a; i < b;)
        if(stack[++ i].real > value.real)
          value.real = stack[i].real;
      break;

    case FN_AVG:                         /* average    */
    case FN_STD:                         /* standard deviation  */
      value.real = 0.0;
      for(i = a; i <= b; i ++)
        value.real += stack[i].real;
      value.real = value.real / (double)params;

      if(code == FN_STD) {
        tmp = 0.0;
        for(i = a; i <= b; i ++)
          tmp += pow(stack[i].real - value.real, 2.0);
        value.real = sqrt(tmp / (double)params);
      }
      break;

    case FN_ROOT:                        /* n:th root of x  */
      if(params == 2)
        value.real = pow(stack[b].real, 1.0 / stack[a].real);
      else
        *err = ARG_ERR;
      break;

    /* k element group combinations in n elements: over(n, k) */
    case FN_OVER:
      if(params == 2)
        value.real = factor(stack[a].real, err) /
                    (factor(stack[b].real, err) *
                     factor(stack[a].real - stack[b].real, err));
      else
        *err = ARG_ERR;
      break;

    /* See the Dr. Dobbs Journal October 1995
     * for floating point -> fraction conversion
     */
    case FN_FRACTION:
      if(params == 2)
        value = tofraction(&stack[a]);
      else
        *err = ARG_ERR;
      break;

    default:
      *err = PRG_ERR;
  }
  if(value.real >= HUGE_VAL || value.imag >= HUGE_VAL)
    *err = OVF_ERR;

  return value;
}

typedef struct
{
  int top;		/* dominator */
  int bot;		/* numerator */
} fraction;

/* converts a floating point number to
 * a fraction of a specified accuracy:
 *     fraction(sqrt(2), 0.00005)
 * ->  239 + 169i
 *    (237 / 169) = 1.4142
 *     sqrt(2)    = 1.41421
 */
static cplx tofraction(cplx stk[])
{
  fraction bot = {0, 1}, top = {1, 1}, med;
  cplx result = {0, 0};
  double diff;
  long ints;

  /* separate the fraction part */
  stk[1].real = (stk[1].real < 0.0 ? -stk[1].real : stk[1].real);
  ints = (int)(stk[0].real > 0.0 ? (floor(stk[0].real)) : (ceil(stk[0].real)));
  stk[0].real = fabs(stk[0].real) - floor(stk[0].real);

  /* check the limits */
  if(stk[0].real == 0.0 || stk[0].real == 1.0)
    return (result.real = (double)ints, result);

  /* loop until fraction is accurate enough */
  do
  {
    med.top = bot.top + top.top;
    med.bot = bot.bot + top.bot;
    diff = (double)med.top / (double)med.bot - stk[0].real;
    if(diff > 0.0)
      top = med;
    else
      bot = med;
    /* check whether enough accuracy or too large fraction */
  } while((diff < 0.0 ? -diff : diff) > stk[1].real && med.bot < (INT_MAX / 2));

  result.real = (double)(med.top + ints * med.bot);
  result.imag = (double)med.bot;
  return result;
}
